------------------------------------------------- 
-- File: mcgenjin.vhd 
-- Entity: mcgenjin
-- Architecture: alteramax
-- Author: Osman Celimli
-- Created: 7/30/2011
-- Modified: 7/31/2011
-- VHDL'93 
-- Description: Memory mapper for PC-Engine.
--   Supports up to 8MB of ROM and internal swapping
--   of the datalines for region control. Two user
--   chipselect lines are given, each with a 256KB
--   mapping area.
--
--   Thus, the 1MB cartslot is remapped as:
--     $00000-$3FFFF Low 256KB of ROM
--     $40000-$7FFFF Selected 256KB ROM Bank
--     $80000-$BFFFF User CS0
--     $C0000-$FFFFF User CS1
--   Writes to even addresses in $00000-$7FFFF set
--   mapper region control. Odd addresses set the
--   current 256KB bank.
--                           76543210
--   Mapper Region Control : xxxxxxxS
--     0 = Data line order normal
--     1 = Data line order reversed
--                           76543210
--   Mapper Bank Control   : xxxBBBBB
--     BBBBB = A22-A18 of the 256KB bank address
--   Note that while reads from ROM are reversed
--     when data line reversal is enabled, writes
--     to the mapper remain unaltered.
-------------------------------------------------  
library ieee;
use ieee.std_logic_1164.all;

entity mcgenjin is Port(
	pce_d8		: inout std_logic_vector(7 downto 0);
	pce_a20,pce_a19	: in std_logic;
	pce_a18,pce_a0	: in std_logic;
	wr,rd,rst	: in std_logic;

	user_cs		: out std_logic_vector(1 downto 0);

	rom_d8		: in std_logic_vector(7 downto 0);
	rom_bank	: out std_logic_vector(4 downto 0);
	rom_ce		: out std_logic
	); 
end entity mcgenjin;

architecture alteramax of mcgenjin is
	--Bidirectional I/O
	component bidir8 is Port(
		inpoutp8	: inout std_logic_vector(7 downto 0);
		tobus8		: in std_logic_vector(7 downto 0);
		frombus8	: out std_logic_vector(7 downto 0);
		oe,swp		: in std_logic
		); 
	end component bidir8;
	--Internal Signals
	signal int_wr,int_rd,int_rst : std_logic;
	signal int_bank : std_logic_vector(4 downto 0);
	signal bidir_oe,bidir_swp : std_logic;
	signal bidir_frombus8,bidir_tobus8 : std_logic_vector(7 downto 0);
begin
	--Map in bidirectional databus with swap to pce
	bidir_map: component bidir8 port map(pce_d8,bidir_tobus8,bidir_frombus8,
					bidir_oe,bidir_swp);
	--Route control signals
	int_wr <= wr;
	int_rd <= rd;
	--ROM DataBus
	bidir_tobus8 <= rom_d8;
	--ROM Bank
	rom_bank <= int_bank;
	--ROM enable only in lower 512KB to save power
	rom_ce <= pce_a20 or pce_a19;
	--User Chipselects
	user_cs(0) <= not((not pce_a20) and pce_a19 and (not pce_a18));
	user_cs(1) <= not((not pce_a20) and pce_a19 and (pce_a18));

	--State Mechanism
	process(int_wr,int_rd,rst)
		variable reg_bank : std_logic_vector(4 downto 0);
		variable reg_swp : std_logic;
	begin
		--Async reset: Bank = 0, Swap = 0
		if(rst = '0') then
			reg_bank := "00000";
			reg_swp := '0';
			--No Dbus assertion
			bidir_oe <= '1';
		--Hu6280 is Writing, Catch for mapper if in lower 512KB
		elsif(int_wr = '0' and pce_a20 ='0' and pce_a19 = '0') then
			--Even addresses go to swap
			if(pce_a0 = '0') then
				reg_swp := bidir_frombus8(0);
			--Odds go to bank
			else
				reg_bank := bidir_frombus8(4 downto 0);
			end if;
			--No Dbus assertion
			bidir_oe <= '1';
		--Hu6280 is Reading, Catch for ROM
		elsif(int_rd = '0' and pce_a20 ='0' and pce_a19 = '0') then
			--Dbus assertion, enable output
			bidir_oe <= '0';
		--Any other case
		else
			--No Dbus assertion
			bidir_oe <= '1';
		end if;

		--No banking in first 256KB
		if(pce_a18 = '0') then
			int_bank <= "00000";
		--Assert Bank Register for top 256KB
		else
			int_bank <= reg_bank;
		end if;
		--Assert Swap Register
		bidir_swp <= reg_swp;
	end process;
end alteramax;

